#!/usr/bin/env python
# encoding: utf-8
"""
Stub class & auto-initialize for all of the various mechanisms that we can use to get data.

Copyright (c) 2015 University of Wisconsin SSEC. All rights reserved.
"""

import os
import glob
import logging
import fcntl

from cspp_fetch.getters import ftp, http, local, rsync
from cspp_fetch.config import MAX_RETRIES

LOG = logging.getLogger(__name__)


def _protocol(remote):
    if remote.startswith("ftp://"):
        return ftp
    elif remote.startswith("http://"):
        return http
    elif remote.startswith("rsync://"):
        return rsync
    elif remote.startswith("/"):
        return local
    else:
        raise RuntimeError("protocol not recognized: %s" % (remote))


def check(remote):
    try:
        return _protocol(remote).check(remote)
    except:
        return []


def verify(remote, local):
    return _protocol(remote).verify(remote, local)


def check_all(remotes):
    for r in remotes:
        if not check(r):
            return False
    return True


# every "get" function should go through this to do the actual download
def download(remote, local):
    return _protocol(remote).download(remote, local)


def get(remote, local):
    localdir, localfile = os.path.split(local)
    if localdir != '' and not os.path.exists(localdir):
        os.makedirs(localdir)

    lock_filename = os.path.join(localdir, "." + localfile + ".lock")
    lock = open(lock_filename, 'a')
    fcntl.lockf(lock, fcntl.LOCK_EX)
    numRetries = 0
    # FIXME: recalling verify from scratch is probably more expensive than
    # we'd like? can we cache something?
    while numRetries < MAX_RETRIES:
        got = download(remote, local)
        if verify(remote, local):
            break
        else:
            LOG.error(
                "[remote]: %s doesn't match [local]: %s, retrying..." % (remote, local))
            numRetries = numRetries + 1

    if verify(remote, local):
        fcntl.lockf(lock, fcntl.LOCK_UN)
        lock.close()
        return got
    else:
        # FIXME: we should download files as hidden and move when they're
        # downloaded, a'la rsync
        os.remove(local)
        fcntl.lockf(lock, fcntl.LOCK_UN)
        lock.close()
        raise RuntimeError("Couldn't sucessfully retrieve %s" % (remote))


def get_any(get_list):
    """
    Goes through a list of (remote, local) and transfers as many files as possible.

    Returns a list of local files successfully transferred.
    """
    got = []
    new_get_list = []
    # check could be replaced by check with slight modification to check
    for remote, local in get_list:
        LOG.debug('remote %s' % remote)
        LOG.debug('local %s' % local)
        checked = check(remote)
        for checkb in checked:
            LOG.debug('record for download ')
            wild_pair = (checkb, os.path.join(local, os.path.basename(checkb)))
            LOG.debug(wild_pair)
            new_get_list.append(wild_pair)

    # LOG.debug('now get')
    # LOG.debug(new_get_list)

    for remote, local in new_get_list:
        get(remote, local)
        got.append(local)

    return got


def get_all(get_list):
    """
    Goes through a list of (remote, local) and, IF AND ONLY IF all of the remotes exist, grabs them all.

    The import difference is that a new get_list is generated by calling check


    Returns a list of local files successfully transferred.
    """
    got = []
    new_get_list = []
    # check could be replaced by check with slight modification to check
    for remote, local in get_list:
        checked = check(remote)
        for checkb in checked:
            new_get_list.append(
                (checkb, os.path.join(local, os.path.basename(checkb))))

    if len(get_list) != len(new_get_list):
        return got

    for remote, local in new_get_list:
        actual = get(remote, local)
        got.append(actual)

    return got


def get_first(get_list, wild=None, how_many=1):
    """
    Goes through a list of (remote, local) and gets the first pair available.

    Returns a list of local files successfully transferred.
    """
    got = []
    new_get_list = []
    # check could be replaced by check with slight modification to check

    for remote, local in get_list:
        local_check = local
 
        #LOG.debug('remote %s' % remote)
        #LOG.debug('local %s ngl %d' % (local,len(new_get_list)))

        if wild is not None:
            local_check = os.path.join(local, wild)
            #LOG.debug('local check %s' % local_check)

        if len(got) < how_many:
            files = check(local_check)
            for f in files:
                got.append(f)
                got = sorted(list(set(got)))

                LOG.debug('Add local got %d how %d file %s' %(len(got) , how_many, f))


            checked = check(remote)
            LOG.debug(checked)
    
            if len(checked) < len(got):
                LOG.debug('Local files not on server or already local len(checked) != len(got) %d %d ' % (
                    len(checked), len(got)))
    
            for checkb in checked:
                wild_pair = (checkb, os.path.join(local, os.path.basename(checkb)))
                new_get_list.append(wild_pair)
                LOG.debug('get new %d  max %d'%(len(new_get_list),how_many))
                LOG.debug('Wild card %s b%s' % (wild_pair))
  


    # want to sort based on date and return most current
    # LOG.info(new_get_list)

    new_get_list = sorted(list(set(new_get_list)))

    # return newest
    for remote, local in reversed(new_get_list):
  
        if len(got) <= how_many:
            LOG.debug('remote %s local %s'%(remote, local))

            msg="downloaded"
            if os.path.exists(local):
                LOG.debug('cached %s'%local)
                msg="cached"
            else :
                get(remote, local)
                LOG.debug('new %s'%local)
 
                
            if os.path.exists(local):
            # we got a file
                if not local in got:
                    got.append(local)
                    LOG.info('Add %s file %s' %(msg, local))

            # return got
    
    return got
