#!/usr/bin/env python
# encoding: utf-8
"""
$Id:$

Purpose: Run the VIIRS SDR using Raytheon ADL 3.1 or 4.0

Input:
    One or more HDF5 VIIRS RDR input files, aggregated or single-granule.
    You need one granule before and one granule after a given granule to
    process. A work directory, typically, empty, in which to unpack the
    granules and generate the output. If the work directory specified does
    not exist, it will be created.

Output:
    ADL VIIRS SDR blob files, .lutasc metadata files, and HDF5 output granules
    will be created.

Details:
    If you input a series of granules, the software will scan the work directory.
    Thus, for N input RDR granules, you may get up to N-2 output SDR granules if they are all contiguous.
    It is ambiguous to provide several copies of the same granule in the work directory;
    this will result in an error abort.
    The unpacker gives each unpacking of a given granule its own
Preconditions:
    Requires ADL_HOME, CSPP_SDR_HOME, CSPP_RT_ANC_CACHE_DIR, CSPP_ANC_HOME  environment variables are set.
    Requires that any needed LD_LIBRARY_PATH is set.
    Requires that DSTATICDATA is set.

Copyright (c) 2011 University of Wisconsin Regents.
Licensed under GNU GPLv3.
"""

import os
import sys
import logging
import glob
import traceback
import time
import datetime as dt
import shutil
import re
from subprocess import CalledProcessError

# skim and convert routines for reading .lutasc metadata fields of interest
# from adl_asc import skim_dir, RDR_REQUIRED_KEYS

from adl_query_dms import skim_dir, setup_dms_in_work_directory,check_spacecraft_diaries

import adl_log

import adl_common
from adl_common import env
from adl_common import status_line, configure_logging, get_return_code
from adl_common import check_env, check_and_convert_path
from adl_common import initialize_work_ADL_variables, ADL_VARS
from adl_common import COMMON_LOG_CHECK_TABLE, check_anc_cache
from adl_common import CSPP_RT_ANC_CACHE_DIR, ADL_HOME
from adl_common import check_existing_env_var, CSPP_RT_HOME, check_static
from adl_post_process import repack_products, aggregate_products_par, add_geo_attribute_to_aggregates
from adl_post_process import SHORTNAME_2_PRODUCTID
from multiprocessing import Pool
import i2m
import rdr_unpack
from sdr_common import check_products, build_workspace, setup_multi_process, trim_the_granules, \
    start_out_good, report_gran_completion, \
    number_good_granules, check_memory, check_number_processors, move_products_to_work_directory
from sdr_common import add_geo_attribute_to_h5
from sdr_common import abs_filenames, rename_products
from adl_viirs_data import ADL_VIIRS_ANC_GLOBS, GIG_MEMORY_PER_PROC, VIIRS_GRAN_ID_DELTA
import adl_anc_retrieval
from sdr_common import cleanup_work_dir
import pprint
pp = pprint.PrettyPrinter(indent=4)
# maximum delay between granule end time and next granule start time to
# consider them contiguous
MAX_CONTIGUOUS_SDR_DELTA = dt.timedelta(seconds=600)

# every module should have a LOG object
LOG = logging.getLogger(__name__)

# keys used in metadata dictionaries
K_FILENAME = '_asc_filename'

ANCILLARY_SUB_DIR = "linked_data"

# locations of executables in ADL 3.1
# and the patterns we're looking for

# CSPP for ADL 4.2
ADL_VIIRS_ANC_BAD_GLOBS = (
    ('VIIRS-SDR-RADIOMETRIC-PARAM-LUT', 'I1.5.05.00'),
    ('VIIRS-SDR-RADIOMETRIC-PARAM-LUT', 'I1.5.06.02'),
    ('*anothr*', 'I1.5.06.02')
)

ADL_VIIRS_GEO_PRODUCT_SHORTNAMES = [
    'VIIRS-MOD-GEO-TC',
    'VIIRS-IMG-GEO-TC',
    'VIIRS-DNB-GEO',
]

OPTIONAL_GEO_PRODUCTS = [
    'VIIRS-IMG-GEO',
    'VIIRS-MOD-GEO'
]

ADL_VIIRS_SDR_PRODUCT_SHORTNAMES = [

    'VIIRS-I1-SDR',
    'VIIRS-I2-SDR',
    'VIIRS-I3-SDR',
    'VIIRS-I4-SDR',
    'VIIRS-I5-SDR',
    'VIIRS-M1-SDR',
    'VIIRS-M2-SDR',
    'VIIRS-M3-SDR',
    'VIIRS-M4-SDR',
    'VIIRS-M5-SDR',
    'VIIRS-M6-SDR',
    'VIIRS-M7-SDR',
    'VIIRS-M8-SDR',
    'VIIRS-M9-SDR',
    'VIIRS-M10-SDR',
    'VIIRS-M11-SDR',
    'VIIRS-M12-SDR',
    'VIIRS-M13-SDR',
    'VIIRS-M14-SDR',
    'VIIRS-M15-SDR',
    'VIIRS-M16-SDR',
    'VIIRS-DNB-SDR'
]

RETAINED_EDR_COLLECTION_REGEXES = [
    'VIIRS-MOD-GEO',
    'VIIRS-IMG-GEO.*',
    'VIIRS-DNB-GEO.*',
    'VIIRS-M\d+-SDR',
    'VIIRS-I\d-SDR',
    'VIIRS-DNB-SDR',
    'VIIRS-M\d+-FSDR',
    'VIIRS-I\d-FSDR',
    'VIIRS-DNB-FSDR',
    'VIIRS-.*?-GRC.*'
]

RETAINED_EDR_COLLECTIONS = map(re.compile, RETAINED_EDR_COLLECTION_REGEXES)

CHECK_REQUIRED_KEYS = ['N_Granule_ID', 'n_collection_short_name']
OBSERVE_TIME = 'ObservedStartTime'

ADL_VIIRS_CAL = os.path.join(ADL_HOME, 'bin', 'ProSdrViirsCalController.exe')

XML_TMPL_VIIRS_CAL = """
<?xml version="1.0" encoding="UTF-8"?>
<!--
  Config file to run a WFM-managed process in lightweight (no WFM) mode.

  Run a WFM-managed process with no args for help on the format of this file.
-->
<InfWfm_TkLwConfig xmlns="http://jpss.lwtkcfg.ray.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://jpss.lwtkcfg.ray.com ExecutionConfig.xsd">
  <processName>ProSdrViirsCalController.exe</processName>
  <perfEnable>FALSE</perfEnable>
  <perfPath>${GRANULE_OUTPUT_DIR}</perfPath>
  <debugDest>D_FILE</debugDest>
  <debugLevel>DBG_HIGH</debugLevel>
  <debugPath>${GRANULE_OUTPUT_DIR}</debugPath>
  <rootDir>${GRANULE_OUTPUT_DIR}</rootDir>
  <startMode>INF_STARTMODE_COLD</startMode>
  <healthInterval>30</healthInterval>
  <configGuideName>ProSdrViirsCalController_GuideList.cfg</configGuideName>
"""

XML_TMPL_VIIRS_CAL_full = """
<?xml version="1.0" encoding="UTF-8"?>
<!--
  Config file to run a WFM-managed process in lightweight (no WFM) mode.

  Run a WFM-managed process with no args for help on the format of this file.
-->
<InfWfm_TkLwConfig xmlns="http://jpss.lwtkcfg.ray.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://jpss.lwtkcfg.ray.com ExecutionConfig.xsd">
  <processName>ProSdrViirsCalController.exe</processName>
  <perfEnable>FALSE</perfEnable>
  <perfPath>${GRANULE_OUTPUT_DIR}</perfPath>
  <debugDest>D_FILE</debugDest>
  <debugLevel>DBG_HIGH</debugLevel>
  <debugPath>${GRANULE_OUTPUT_DIR}</debugPath>
  <rootDir>${GRANULE_OUTPUT_DIR}</rootDir>
  <startMode>INF_STARTMODE_COLD</startMode>
  <healthInterval>30</healthInterval>
  <configGuideName>ProSdrViirsCalController_GuideList.cfg</configGuideName>

  <task>
    <taskType>GRANULE</taskType>
    <granuleId>%(N_Granule_ID)s</granuleId>
    <granuleVersion>%(N_Granule_Version)s</granuleVersion>
    <spacecraft>${SPCRFT}</spacecraft>
    <sensor>VIIRS</sensor>
  </task>
  <task>
    <taskType>SHUTDOWN</taskType>
  </task>
</InfWfm_TkLwConfig>
"""


ADL_VIIRS_GEO = os.path.join(ADL_HOME, 'bin', 'ProSdrViirsGeoController.exe')
XML_TMPL_VIIRS_GEO = """
<?xml version="1.0" encoding="UTF-8"?>
<!--
  Config file to run a WFM-managed process in lightweight (no WFM) mode.

  Run a WFM-managed process with no args for help on the format of this file.
-->
<InfWfm_TkLwConfig xmlns="http://jpss.lwtkcfg.ray.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://jpss.lwtkcfg.ray.com ExecutionConfig.xsd">
  <processName>ProSdrViirsGeoController.exe</processName>
  <perfEnable>FALSE</perfEnable>
  <perfPath>${GRANULE_OUTPUT_DIR}</perfPath>
  <debugDest>D_FILE</debugDest>
  <debugLevel>DBG_HIGH</debugLevel>
  <debugPath>${GRANULE_OUTPUT_DIR}</debugPath>
  <rootDir>${GRANULE_OUTPUT_DIR}</rootDir>
  <startMode>INF_STARTMODE_COLD</startMode>
  <healthInterval>30</healthInterval>
  <configGuideName>ProSdrViirsGeoController_GuideList.cfg</configGuideName>
"""
XML_TMPL_VIIRS_GEO_full = """
<?xml version="1.0" encoding="UTF-8"?>
<!--
  Config file to run a WFM-managed process in lightweight (no WFM) mode.

  Run a WFM-managed process with no args for help on the format of this file.
-->
<InfWfm_TkLwConfig xmlns="http://jpss.lwtkcfg.ray.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://jpss.lwtkcfg.ray.com ExecutionConfig.xsd">
  <processName>ProSdrViirsGeoController.exe</processName>
  <perfEnable>FALSE</perfEnable>
  <perfPath>${GRANULE_OUTPUT_DIR}</perfPath>
  <debugDest>D_FILE</debugDest>
  <debugLevel>DBG_HIGH</debugLevel>
  <debugPath>${GRANULE_OUTPUT_DIR}</debugPath>
  <rootDir>${GRANULE_OUTPUT_DIR}</rootDir>
  <startMode>INF_STARTMODE_COLD</startMode>
  <healthInterval>30</healthInterval>
  <configGuideName>ProSdrViirsGeoController_GuideList.cfg</configGuideName>

  <task>
    <taskType>GRANULE</taskType>
    <granuleId>%(N_Granule_ID)s</granuleId>
    <granuleVersion>%(N_Granule_Version)s</granuleVersion>
    <spacecraft>${SPCRFT}</spacecraft>
    <sensor>VIIRS</sensor>
  </task>
  <task>
    <taskType>SHUTDOWN</taskType>
  </task>
</InfWfm_TkLwConfig>
"""



def sift_metadata_for_run_all_granules(spcrft, work_dir='.'):
    """
    search through the ASC metadata in a directory,
    grouping in StartTime order
    look for back-to-back granules and break into contiguous sequences
    check that S/C diary RDRs are available for granules of interest
    yield sequence of granules we can process
    """

    def start_time_key(x): return (
        x['StartTime'], x.get('N_Granule_Version', None))

    gs = sorted(skim_dir(n_collection_short_name='SPACECRAFT-DIARY-RDR',
                         spcrft=spcrft),
                key=start_time_key)

    # eliminate_duplicates(gs,
    #                     larger_granules_preferred=True)

    LOG.debug('sifting science RDR data for processing opportunities')

    gs = sorted(skim_dir(n_collection_short_name='VIIRS-SCIENCE-RDR',
                         spcrft=spcrft),
                key=start_time_key)

    # eliminate_duplicates(gs)
    Viirs_Science_RDRs = list(
        skim_dir(n_collection_short_name="VIIRS-SCIENCE-RDR", spcrft=spcrft))
    


    status_line("Total Viirs Science RDRs: " + str(len(Viirs_Science_RDRs)))

    def start_time_key(x): return (
        x['StartTime'], x.get('N_Granule_Version', None))

    for group in sorted(Viirs_Science_RDRs, key=start_time_key):
        yield group


# Description: Required input not available for Shortname

# table of ADL LOG error messages and the associated hint to correcting
# the problem
viirs_sdr_log_check_table = [
    ("Verified RDR has invalid mode for geolocation/calibration",
     "Likely bad data at start or end of pass "),
    ("PRO_FAIL Unable to read",
     "Check CSPP/static and CSPP/cache directories"),
    ("Required input not available",
     "Missing or out of date ancillary input, Check effectivity for listed Shortname."),
    ("PRO_FAIL runAlgorithm()", "Algorithm failed"),
    ("Completed unsuccessfully", "Algorithm failed"),
    ("arbitrary time is invalid",
     "Problem with input RDR,check NPP_GRANULE_ID_BASETIME"),
    ("Error retrieving data for USNO-POLARWANDER-UT1",
     "POLAR WANDER file needs update,check NPP_GRANULE_ID_BASETIME"),
    ("Algorithm failed", "Controller did not run, check log")
]


# Look through new log files for completed messages
def checkADLLogForSuccess(work_dir, pid, xml):
    """
    Find the log file
    Look for success
    Return True if found
    Display log message and hint if problem occurred
    """

    logDir = work_dir

    files = glob.glob(os.path.join(logDir, "*.log"))
    status_line(
        "Checking " + str(len(files)) + " log files for errors" + logDir)
    logFile = "No_log"
    n_err = 0
    for logFile in files:
        LOG.info("Checking Log file " + logFile + " for errors.")
        count = adl_log.scan_log_file(
            COMMON_LOG_CHECK_TABLE + viirs_sdr_log_check_table, logFile)

        n_err += count

    if n_err == 0:
        status_line("Processing of file: " + xml + " Completed successfully")
        return True
    else:
        status_line("Processing of file: " + xml +
                    " Completed unsuccessfully, Look at previous message")
        LOG.debug("Log file: " + logFile)
        return False


def _skim_dir_collections(*args, **kwargs):
    """skim and return dictionary keyed on n_collection_short_name,
    each entry being a list of skimmed info dictionaries
    """
    from collections import defaultdict

    zult = defaultdict(list)
    for file_info in skim_dir(*args, **kwargs):
        key = file_info.get('n_collection_short_name', '-unknown-collection-')
        zult[key].append(file_info)
    return zult


def move_retained_edr_blobs(path, dry_run=False):
    """
    Move the ip blobs and lutasc files to work directory
    These files may be used in later processing
    """
    LOG.debug("move retained ip")
    collections = _skim_dir_collections(path)
    files = []

    def of_interest(cn):
        for regex in RETAINED_EDR_COLLECTIONS:
            if regex.match(cn):
                return True
        return False

    for cn, ascs in collections.items():
        if not of_interest(cn):
            LOG.debug('ignoring collection %s' % cn)
            continue
        LOG.debug('%d blobs for %s' % (len(ascs), cn))
        for asc in ascs:
            # LOG.debug('keys %s' % repr(list(lutasc.keys()))) # FIXME
            blob = asc.get('BlobPath', '')
            # not always the case that every lutasc has a blob (night scenes)
            if os.path.isfile(blob):
                files.append(blob)
            # FUTURE: K_FILENAME different from adl_asc.K_FILENAME??
            files.append(asc['_filename'])

    dest = os.path.dirname(
        path)  # FUTURE this assumes that the work directory is always a subdir of the target directory
    LOG.debug('%d edr product files to move to %s' % (len(files), dest))

    for blob_file in files:
        name = os.path.basename(blob_file)
        if not dry_run:
            shutil.move(os.path.join(path, name), dest)
        else:
            LOG.debug("mv %s %s" % (os.path.join(path, name), dest))


def find_granules_and_build_xml(work_dir, spcrft):
    # read through ascii metadata and build up information table
    start_time = time.time()
    LOG.debug("<< find_granules_and_build_xml  >>")

    granules_to_process = []
    for gran in sift_metadata_for_run_all_granules(spcrft, work_dir):
        granules_to_process.append(gran)

    # diaries = list(contiguous_granule_groups(
    # skim_dir(work_dir, required_keys=RDR_REQUIRED_KEYS, n_collection_short_name='SPACECRAFT-DIARY-RDR'),
    # larger_granules_preferred=True))

    LOG.debug(', '.join(x['N_Granule_ID'] for x in granules_to_process))
    if not granules_to_process:
        LOG.error("Found no granules to process!")
        return []
    else:
        LOG.info("Found %d granules to process." % (len(granules_to_process)))

    end_time = time.time()
    elapsed_time = end_time - start_time
    LOG.debug("<< Find Granules Execution Time: %f >>" % elapsed_time)

    return granules_to_process


def reconstruct(granule_output_dir):
    """
    reconstruct SVM07 from SVI02 if SVM07 is all missing
    """
    svi02_files = glob.glob(os.path.join(granule_output_dir, "SVI02*"))

    for svi02_file in svi02_files:
        temp = os.path.basename(svi02_file).split("_c")
        svm_file = temp[0]
        svm_file = svm_file.replace("SVI02", "SVM07")
        svm_file = svm_file + "*.h5"
        svm_file = glob.glob(os.path.join(granule_output_dir, svm_file))
        if len(svm_file) > 0:
            LOG.debug("File pair %s %s" % (svi02_file, svm_file[0]))

            i2m.i2m(svi02_file, svm_file[0], FORCE=False)


def run_and_check_single_granule(additional_env):
    """run a VIIRS SDR XML input in sequence"""

    work_dir = additional_env['WORK_DIR']
    xml_files_str = additional_env['XML_FILE']
    gran_id = additional_env['N_Granule_ID']
    granule_output_dir = additional_env['GRANULE_OUTPUT_DIR']
    compress = additional_env['COMPRESS']
    cleanup = additional_env['CLEANUP']
    edr = additional_env['EDR']
    chunk_size=additional_env['CHUNK_SIZE']
    chunks = 0
   

    start_time = time.time()
    LOG.debug("<< run_and_check_single_granule  >>")

    error_files = []
    xml_files = xml_files_str.split(":")

    good = True

    try:
        f = 0
        for xml_file in xml_files:
            if f == 0:
                cmd = [ADL_VIIRS_GEO, xml_file] 
            elif f == 1:
                cmd = [ADL_VIIRS_CAL, xml_file]

            status_line('Executing %r with WORK_DIR=%r ' % (cmd, work_dir))
            LOG.debug('additional environment vars: %s' %
                      repr(additional_env))
            z = ADL_VARS.copy()
            # which returns None since it mutates z
            z.update(additional_env)

            LOG.debug(pp.pformat(z))

            pid = adl_common.sh(cmd, env=env(**z), cwd=work_dir)
            #cmd = " ".join(cmd)
            #cmd = "/usr/bin/strace %s"%cmd
            #pid = adl_common.execute_binary_captured_io(work_dir, cmd, log_execution=True, **z)
            
            LOG.debug("%r ran as pid %s" % (cmd, str(pid)))

            ran_ok = checkADLLogForSuccess(
                     granule_output_dir, pid, xml_file)
            f += 1

    except CalledProcessError as oops:

        pid = getattr(oops, 'pid', None)
        ran_ok = checkADLLogForSuccess(granule_output_dir, pid, xml_file)
        LOG.debug(traceback.format_exc())
        LOG.error(
            'ProSdrViirsController.exe failed on %r: %r. Continuing...' % (xml_file, oops))
        error_files.append(xml_file)

    try:
        if error_files:
            LOG.warning('Had problems running these XML files: %r' %
                        (error_files,))
            good = False
        # else:
        if True:
            LOG.debug(
                "Checking that output granule blobs exist for granule: " + gran_id)
            #   Check For Products/ Errors

            product_shortnames = sorted(
                ADL_VIIRS_SDR_PRODUCT_SHORTNAMES) + sorted(ADL_VIIRS_GEO_PRODUCT_SHORTNAMES)
            problem = check_products(
                granule_output_dir, gran_id, product_shortnames,chunk_size=chunk_size)

            if problem is True:
                good = False

            if good is True:
                # after run check results and clean up
                #   PATCH THE OUTPUTS
                product_ids = []

                to_move = ADL_VIIRS_SDR_PRODUCT_SHORTNAMES + \
                    ADL_VIIRS_GEO_PRODUCT_SHORTNAMES
                for short_name in to_move:
                    sdr_name = SHORTNAME_2_PRODUCTID[short_name]
                    product_ids.append(sdr_name)
                rename_products(granule_output_dir, product_ids)

                add_geo_attribute_to_h5(
                    granule_output_dir, gran_id, ADL_VIIRS_SDR_PRODUCT_SHORTNAMES)

                # reconstruct SVM07 from SVI02 if SVM07 is missing
                #if additional_env['SPCRFT'] == 'NPP' :
                reconstruct(granule_output_dir)

                if compress == "True":
                    status_line('Compress products for %s' % gran_id)
                    repack_products(
                        granule_output_dir, ADL_VIIRS_SDR_PRODUCT_SHORTNAMES)
                    repack_products(
                        granule_output_dir, ADL_VIIRS_GEO_PRODUCT_SHORTNAMES)
                    status_line('Compress products for %s complete.' % gran_id)

                if edr == "True":
                    move_retained_edr_blobs(granule_output_dir)

                chunks , option_chunks = move_products_to_work_directory(
                    granule_output_dir, product_ids)

                if cleanup == 'True':
                    LOG.debug("Remove: " + granule_output_dir)
                    if os.path.exists(granule_output_dir):

                        removable_files = glob.glob(
                            os.path.join(granule_output_dir, "*"))
                        for remove_file in removable_files:
                            if os.path.isfile(remove_file):
                                os.remove(remove_file)
                            if os.path.isdir(remove_file):
                                shutil.rmtree(remove_file)
                    else:
                        LOG.debug("Unable to remove:" + granule_output_dir)
    except:
        LOG.error("Internal problem for granule %s" % gran_id)
        LOG.error(traceback.format_exc())
        good = False

    end_time = time.time()
    elapsed_time = end_time - start_time
    status_line('Tot Execution Time:  "%f" Gran "%s"' %
                (elapsed_time, gran_id))

    stat = ""
    if good is True:
        stat = 'Granule %s Completed without problems. ' % gran_id

    if good is False:
        stat = 'Granule %s Completed with problems. ' % gran_id

    report_gran_completion(stat,chunks,int(additional_env["ACTUAL"]))

    result = {"GOOD": str(good)}

    return result


def run_all_granules(work_dir, h5_names, trim=None, setup_only=False, cleanup=True, allow_cache_update=True,
                     compress=False, nprocs="1", edr=False, spcrft='NPP',accuracyMode='reduced',center=False, aggregate=False):
    "process VIIRS RDR data in HDF5 format (any aggregation) to SDRs in an arbitrary work directory"
    if not trim:
        trim = [0, 0]
    # ' Toggle to delete all files except h5 after execution'

    noncritical_problem = 0
    cleanup_after_running = cleanup
    start_time = time.time()

    # "Create the working directory and a subdirectory for the logs"
    build_workspace(work_dir)

    status_line("Unpack the supplied inputs")

    num_unpacking_problems = rdr_unpack.unpack_rdr(
        work_dir, h5_names, allow_cache_update=allow_cache_update)
    
    
    check_spacecraft_diaries(spcrft=spcrft)


    status_line("Search through the inputs for legal granule combinations")
    granules_to_process = find_granules_and_build_xml(work_dir, spcrft)
    
    # test mode to pick center granules for shortened test cycle will result in one  granule processed
    if center is True: 
        
        # 10 11
        c = int(len(granules_to_process)/2)
        r = int(len(granules_to_process) % 2)
        trim[0] = c
        trim[1] =  ( c - 1  + r )
        #LOG.info(' r %d c %d o %d '%(r,c,(len(granules_to_process) - ( c - r ))))
        LOG.debug(trim)
    
    #   EVALUATE AND RETRIEVE ANCILLARY DATA
    status_line("Link the required ancillary data into the workspace")

    search_dirs = [CSPP_RT_ANC_CACHE_DIR]

    environment_error = False
    try:

        if len(granules_to_process) > 0:
            # get list of dynamic ancillary files.  Servicing may pull files
            # from remote server.

            LOG.info("Updating ancillary cache")

            dynamic_ancillary_file_list = adl_anc_retrieval.service_remote_ancillary(
                work_dir, granules_to_process,
                adl_anc_retrieval.kRSB, spcrft,
                allow_cache_update=allow_cache_update)

            LOG.info('auto cal done')
            if len(dynamic_ancillary_file_list) <= 0:
                status_line("No RSB autocal use static Mode")
                ADL_VIIRS_ANC_GLOBS.append(
                    '*VIIRS-SDR-CAL-AUTOMATE-LUT*001-static*')
            else:
                ADL_VIIRS_ANC_GLOBS.append(
                    '*VIIRS-SDR-CAL-AUTOMATE-LUT*[0-9]-PE*')
                status_line("RSB autocal enabled")
                LOG.debug(dynamic_ancillary_file_list)

            adl_anc_retrieval.service_remote_ancillary(
                work_dir, granules_to_process,
                adl_anc_retrieval.kPOLAR, spcrft,
                allow_cache_update=allow_cache_update)

            adl_anc_retrieval.service_remote_ancillary(
                work_dir, granules_to_process,
                adl_anc_retrieval.kTLE, spcrft,
                allow_cache_update=allow_cache_update)
            # now update orbit number?
            adl_anc_retrieval.update_orbit_in_db(
                work_dir, granules_to_process, spcrft)

    except EnvironmentError:
        environment_error = True
        LOG.error(traceback.format_exc())
    # RUN THE VIIRS SDR
    granules_to_process = trim_the_granules(
        granules_to_process, VIIRS_GRAN_ID_DELTA, trim=trim)
    total_granules = len(granules_to_process)
    if not environment_error:

        fnxml_geo = 'sdr_viirs_geo_%s.xml'
        fnxml_cal = 'sdr_viirs_cal_%s.xml'

        # % name XML_TMPL_VIIRS_SDR
        chunk_size= int( len(granules_to_process) / int (nprocs) )
        if chunk_size <= 0:
            chunk_size=1
        chunk_size = 1    

        additional_keys = dict(EDR=str(edr), SPCRFT=spcrft,INSTRUMENT='VIIRS',CHUNK_SIZE=str(chunk_size))
        argument_dictionaries, granules_to_process = setup_multi_process(work_dir,
                                                                         granules_to_process,
                                                                         [fnxml_geo,
                                                                             fnxml_cal],
                                                                         [XML_TMPL_VIIRS_GEO,
                                                                          XML_TMPL_VIIRS_CAL],
                                                                         additional_keys,
                                                                         compress=compress,
                                                                         cleanup=cleanup,
                                                                         accuracyMode=accuracyMode)

        total_granules = len(granules_to_process)
        start_out_good(total_granules)

        for gran in granules_to_process:
            gran_id = gran['N_Granule_ID']
            LOG.info('N_Granule_ID %s' % gran_id)

        keep_all_output = False
        results = []

        end_time = time.time()
        elapsed_time = end_time - start_time
        status_line('Setup Execution Time:  "%f" ' % (elapsed_time))

        if len(argument_dictionaries) > 0:
            nprocs = check_number_processors(int(nprocs))
            gig_recomended = GIG_MEMORY_PER_PROC
            check_memory(nprocs, gig_recomended)

            parallel = Pool(int(nprocs))

            try:
                #results = parallel.map_async(run_and_check_single_granule,
                #                             argument_dictionaries).get(9999999)
                results = parallel.map(run_and_check_single_granule,
                                             argument_dictionaries)
                              
            except KeyboardInterrupt:
                LOG.error("Got exception, stopping workers and exiting.")
                parallel.terminate()
                parallel.join()
                sys.exit(1)

        for result in results[1:-1]:
            # do not fail on first or last granule - these can have errors
            if result["GOOD"] != "True":
                LOG.debug("result check %s" % result["GOOD"])
                keep_all_output = True
                noncritical_problem += 1

        LOG.debug("After multi: cleanup %s keep %s" % (str(cleanup_after_running),
                                                       str(keep_all_output)))
        if cleanup_after_running is True and keep_all_output is False:
            for args in argument_dictionaries:
                granule_dir = args['GRANULE_OUTPUT_DIR']
                LOG.debug("Remove: " + granule_dir)
                if os.path.exists(granule_dir):
                    shutil.rmtree(granule_dir)

                    # clean up the science data only if you there are no other
                    # granule directories left

            cleanup_work_dir(work_dir,spcrft=spcrft)

    LOG.debug(pp.pformat(aggregate))
   
    if aggregate is True and keep_all_output is False:
        number_problems = aggregate_products_par(
            work_dir,
            ADL_VIIRS_GEO_PRODUCT_SHORTNAMES,
            nprocs=nprocs,spcrft=spcrft)
        
        number_problems += aggregate_products_par(
            work_dir,
            ADL_VIIRS_SDR_PRODUCT_SHORTNAMES,
            nprocs=nprocs,spcrft=spcrft)
        
        
        product_ids = []

        to_move = ADL_VIIRS_SDR_PRODUCT_SHORTNAMES + \
                    ADL_VIIRS_GEO_PRODUCT_SHORTNAMES
        for short_name in to_move:
            sdr_name = SHORTNAME_2_PRODUCTID[short_name]
            product_ids.append(sdr_name)
            
        rename_products(work_dir,product_ids)
        
        number_problems += add_geo_attribute_to_aggregates(
            work_dir,
            ADL_VIIRS_SDR_PRODUCT_SHORTNAMES)
        
        if number_problems > 0:
            LOG.info('Number aggregation problems %d' % int(number_problems))
            rc = 1



    gran_failure = total_granules - number_good_granules()
    rc = get_return_code(num_unpacking_problems, total_granules,
                         gran_failure, noncritical_problem, environment_error)
    return rc


def viirs_sdr_run(work_dir, h5_names, trim=None, compress=False, aggregate=False, all_geo=False, cleanup=True,
                  allow_cache_update=True, nprocs="1", edr=False, spcrft='NPP',accuracyMode='reduced',center=False):
    """
    callable entry put for running viirs SDR
    @param work_dir:
    @param h5_names:
    @param compress:
    @param aggregate:
    @param all_geo:
    @param setup_only:
    @param cleanup:
    @param allow_cache_update:
    @param trim_granules:
    @param nprocs:
    @param edr:
    """
    if trim is None:
        trim = [0, 0]

    check_env()

    if all_geo is True:
        ADL_VIIRS_GEO_PRODUCT_SHORTNAMES.extend(OPTIONAL_GEO_PRODUCTS)

    LOG.info("allow_cache_update %s" % str(allow_cache_update))
    rc = run_all_granules(
        work_dir,
        h5_names,
        trim=trim,
        cleanup=cleanup,
        allow_cache_update=allow_cache_update,
        compress=compress, nprocs=nprocs, edr=edr, spcrft=spcrft,
        accuracyMode=accuracyMode,center=center,aggregate=aggregate)

#     LOG.debug(pp.pformat(aggregate))
#     LOG.debug(pp.pformat(rc))
#     if aggregate is True and rc == 0:
#         number_problems = aggregate_products_par(
#             work_dir,
#             ADL_VIIRS_GEO_PRODUCT_SHORTNAMES,
#             nprocs=nprocs,spcrft=spcrft)
#         
#         number_problems += aggregate_products_par(
#             work_dir,
#             ADL_VIIRS_SDR_PRODUCT_SHORTNAMES,
#             nprocs=nprocs,spcrft=spcrft)
#         
#         
#         product_ids = []
# 
#         to_move = ADL_VIIRS_SDR_PRODUCT_SHORTNAMES + \
#                     ADL_VIIRS_GEO_PRODUCT_SHORTNAMES
#         for short_name in to_move:
#             sdr_name = SHORTNAME_2_PRODUCTID[short_name]
#             product_ids.append(sdr_name)
#             
#         rename_products(work_dir,product_ids)
#         
#         number_problems += add_geo_attribute_to_aggregates(
#             work_dir,
#             ADL_VIIRS_SDR_PRODUCT_SHORTNAMES)
#         
#         
#         if number_problems > 0:
#             LOG.info('Number aggregation problems %d' % int(number_problems))
#             rc = 1

    return rc


def main():
    """ Run Viirs SDR proccessing
    """
    import argparse

    desc = """Build VIIRS SDR ProSdrViirsDBController work directory and run VIIRS SDR.
        viirs_sdr.sh wraps the python code, adl_viirs_sdr.py options are passed through."""
    parser = argparse.ArgumentParser(description=desc)

    parser.add_argument('-W', '--work-dir', metavar='work_dir', default='.',
                        help='directory in which all activity occurs, defaults to current dir')

    parser.add_argument('-d', '--debug',
                        action="store_true", default=False, help="always retain intermediate files")

    parser.add_argument('-t', '--trim', nargs=2, type=int, metavar='N',
                        default=None,
                        help='Trim granules. -t \"Trim Begin\" \"Trim End\" -t 1 1 Trims first and last granule.')

    parser.add_argument('-z', '--zip',
                        action="store_true", default=False, help="compress products with h5repack zip compression")

    parser.add_argument('--center', dest='center', action='store_true',
                        help='Trim to center granule')
    parser.add_argument('--no-center', dest='center', action='store_false',
                        help='Normal output')
    parser.set_defaults(center=False)
    
    parser.add_argument('-m', '--accuracymode',
                        action="store_true", default=False, help="increase geo accuracy")


    parser.add_argument('-g', '--geo',
                        action="store_true", default=False, help="Retain MOD-GEO products")

#     parser.add_argument('-e', '--edr',
#                         action="store_true", default=False, help="Retain blobs and lutasc for EDR processing")

    parser.add_argument('-a', '--aggregate',
                        action="store_true", default=False, help="aggregate products with nagg")

    parser.add_argument('-l', '--local',
                        action="store_true", default=False, help="disable download of remote ancillary data to cache")

    parser.add_argument('-p', '--processor',
                        type=int, default=1, help="Number of processors to use for granule processing")

    parser.add_argument('-v', '--verbosity', action="count", default=0,
                        help='each occurrence increases verbosity 1 level through ERROR-WARNING-INFO-DEBUG')

    parser.add_argument('filenames', metavar='filename', type=str, nargs='+',
                        help='HDF5 VIIRS RDR file/s to process')

    args = parser.parse_args()

    levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
    level = levels[args.verbosity if args.verbosity < 4 else 3]

    docleanup = True

    work_dir = check_and_convert_path("WORK_DIR", args.work_dir)

#     MAXWORKLENGTH=100
#     if len( work_dir) > MAXWORKLENGTH :
#         LOG.error ("Work directory path too long")
#         sys.exit ( 1 )
        
    d = dt.datetime.now()
    timestamp = d.isoformat()
    logname = "viirs_sdr." + timestamp + ".log"
    logfile = os.path.join(work_dir, logname)
    configure_logging(level, FILE=logfile)
    if args.debug is True:
        docleanup = False

    dpe_version = check_existing_env_var('DPE_VER', default_value=None)
    LOG.info("product version: DPE_VER for VIIRS SDR  %s" % dpe_version)

    check_anc_cache(check_write=not args.local)
    check_static()
    full_db_name = setup_dms_in_work_directory(work_dir, CSPP_RT_HOME)
    initialize_work_ADL_variables(work_dir, full_db_name)

    compress = False
    if args.zip is True:
        compress = True

    accuracyMode='reduced'
    if args.accuracymode is True:
        accuracyMode = 'full'

    aggregate = False
    if args.aggregate is True:
#         if len(args.filenames) > 1:
#             LOG.error("Only one input allowed with aggregate option")
#             sys.exit(2)
        aggregate = True

    LOG.info('CSPP execution work directory is %r' % work_dir)

    if not args.trim:
        args.trim = [0, 0]

    if not args:
        parser.print_help()
        return 1

    rc = 1
    try:
        if 'npp' in os.path.basename( args.filenames[0] ) :
            spcrft = 'NPP'
        else:
            spcrft = 'J01'

        LOG.debug('Spacecraft %s' % spcrft)
        filenames = abs_filenames(args.filenames)

        rc = viirs_sdr_run(work_dir, filenames, trim=args.trim, compress=compress, aggregate=aggregate,
                           all_geo=args.geo,
                           cleanup=docleanup, allow_cache_update=not args.local, nprocs=str(args.processor), 
                           spcrft=spcrft,accuracyMode=accuracyMode,center=args.center)
                           #edr=args.edr,

        if rc == 0 and not args.debug:
            os.remove(logfile)
            cleanup_work_dir(work_dir,spcrft=spcrft,force_all=True)


    except Exception:
        LOG.error(traceback.format_exc())

    return rc


if __name__ == '__main__':
    sys.exit(main())
