#!/usr/bin/env python
# encoding: utf-8

import os
import sys
import logging
import re
import glob
import shutil

from adl_common import status_line
from adl_common import ADL_HOME,  ADL_VARS
from multiprocessing import Lock, Value, cpu_count
import traceback
import time
import adl_geo_ref
import h5py
import datetime as dt

from adl_post_process import SHORTNAME_2_PRODUCTID
from adl_geo_ref import GEO_GUIDE
import pprint


pp = pprint.PrettyPrinter(indent=4)
LOG = logging.getLogger(__name__)


XML_TMPL_START = """
<?xml version="1.0" encoding="UTF-8"?>
<!--
  Config file to run a WFM-managed process in lightweight (no WFM) mode.

  Run a WFM-managed process with no args for help on the format of this file.
-->
<InfWfm_TkLwConfig xmlns="http://jpss.lwtkcfg.ray.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://jpss.lwtkcfg.ray.com ExecutionConfig.xsd">
  <processName>ProSdrCrisFSController.exe</processName>
  <perfEnable>FALSE</perfEnable>
  <perfPath>${GRANULE_OUTPUT_DIR}</perfPath>
  <debugDest>D_FILE</debugDest>
  <debugLevel>DBG_HIGH</debugLevel>
  <debugPath>${GRANULE_OUTPUT_DIR}</debugPath>
  <rootDir>${GRANULE_OUTPUT_DIR}</rootDir>
  <startMode>INF_STARTMODE_COLD</startMode>
  <healthInterval>30</healthInterval>
  <configGuideName>ProSdrCrisFSController_GuideList.cfg</configGuideName>
"""
XML_TMPL_TASK = """
  <task>
    <taskType>GRANULE</taskType>
    <granuleId>%(N_Granule_ID)s</granuleId>
    <granuleVersion>A1</granuleVersion>
    <spacecraft>%(SPCCRFT)s</spacecraft>
    <sensor>%(INSTRUMENT)s</sensor>
  </task>
"""

XML_TMPL_END = """
  <task>
    <taskType>SHUTDOWN</taskType>
  </task>
</InfWfm_TkLwConfig>
"""

def generate_sdr_xml2(work_dir, granule_chunk, controller_name, controller_template,instrument,spccrft):
    gran = granule_chunk[0]
    name = gran['N_Granule_ID']

    fnxml = controller_name % name
    fpxml = open(os.path.join(work_dir, fnxml), 'wt')
    # start
    fpxml.write(controller_template)
    # task
    for gran in granule_chunk:
        gran['INSTRUMENT']=instrument
        gran['SPCCRFT']=spccrft
        fpxml.write(XML_TMPL_TASK % (gran))
 
    #end
    fpxml.write(XML_TMPL_END)
    
    status_line('Created ADL controller XML %s for %s' %
                (fnxml, gran['N_Granule_ID']))

    return fnxml

def add_geo_attribute_to_h5(work_dir, gran_id, shortnames):
    """ Adds GEO_REF property to produced granules
    """
    added = 0

    LOG.info("Add GEO_REF property to : " + gran_id + " products.")

    for short_name in sorted(shortnames):

        prod_id = SHORTNAME_2_PRODUCTID[short_name]
        
        LOG.debug('look %s' % os.path.join(work_dir, '%s*.h5' % prod_id))
        gran_list = glob.glob(os.path.join(work_dir, '%s*' % prod_id))

        for pat, stem, refid in GEO_GUIDE:
            if pat in prod_id:
                geo_prod = stem

        #if len(gran_list) == 1:
        for sdr_name in gran_list:
                try:
                    sname = os.path.basename(sdr_name).split("_")
                    LOG.debug("GEO Product: " + sdr_name + " produced.")
                    LOG.debug(
                        "Oberserved : " + sdr_name + " short_name: " + short_name)
                    if geo_prod is not None:
                        geo_name = geo_prod + "*%s*.h5"%sname[3]
                        gnames = glob.glob(
                            os.path.join(work_dir, geo_name))

                        if len(gnames) > 0:
                            geo_filename = gnames[0]

                            adl_geo_ref.write_geo_ref(
                                sdr_name, geo_filename=geo_filename)
                            added += 1
                except KeyError:
                    LOG.debug("Key error on blob property " + 'ObservedStartTime' +
                              " or " + short_name)
    if added > 0:
        LOG.info("Added N_GEO_Ref properties to " + str(added) +
                 " files for " + gran_id + " products")


def rename_products(work_dir, product_ids):
    """ Checks that proper products were produced.
    If product h5 file was produced blob and lutasc files are deleted
    If product was not produced files are left alone
    """
    #dest = os.path.dirname(work_dir)
    good = True
    for product_id in product_ids:
        files = glob.glob(os.path.join(work_dir, '%s*.h5' % product_id))
     
        for sdrproduct in files:

            if sdrproduct.find("_d1958") != -1:
                LOG.warn("IGNORE File with bad year: " + sdrproduct)
            else:
                new_name = os.path.basename(
                    sdrproduct.replace('_cspu_', '_cspp_'))
                new_name = os.path.basename(
                    new_name.replace('_pop', '_dev'))
                new_name = os.path.basename(
                    new_name.replace('_ops', '_dev'))
                
                #new_name = os.path.basename(
                 #   new_name.replace('npp', spcrft))

              
                shutil.move(sdrproduct, os.path.join(work_dir, new_name))
                LOG.debug('Product %s ' % sdrproduct)

        if len(files) != 1:
            LOG.debug('Product %s missing. %d products produced.' %
                      (product_id, len(files)))
            good = False

    return good

#Beginning_Time Ending_Time
def h5_urid(pathname, attrpath):
    "obtain the N_Reference_ID attribute from an XDR or GEO file"
    h5 = h5py.File(pathname, 'r')
    hob = h5
    x = attrpath.split('/')
    for oname in x[:-1]:
        hob = hob[oname]
    urid = hob.attrs[x[-1]][0][0]
    h5.close()
    return urid


#VIIRS-M16-SDR_Gran_0 VIIRS-M16-SDR_Aggr VIIRS-M16-SDR Data_Products
def obs_dts(gfile ):
    n=gfile.split('_')
    
    tf='%s:%s'%(n[2][1:],n[3][1:])

    obs_start_dt = dt.datetime.strptime(tf,'%Y%m%d:%H%M%S%f')
    tf='%s:%s'%(n[2][1:],n[4][1:])
    obs_end_dt = dt.datetime.strptime( tf ,'%Y%m%d:%H%M%S%f')
    

    return obs_start_dt , obs_end_dt

# def millis_interval(start, end):
#     """start and end are datetime instances"""
#     diff = end - start
#     millis = diff.days * 24 * 60 * 60 * 1000
#     millis += diff.seconds * 1000
#     millis += diff.microseconds / 1000
#     return millis


# def eval_start_and_end(work_dir, gfile):
#     
#     istart, gran_end_time, calc_gran_time, gran_start_dt, gran_end_dt  = granid(work_dir,os.path.basename(work_dir))
#     obs_start_dt, obs_end_dt = obs_dts(gfile)
#     end_delta = gran_end_dt - obs_end_dt    
#         
#     emi = millis_interval( gran_end_dt , obs_end_dt ) 
#     msec = calc_gran_time  - abs (emi * 1000)
#     # 84100000   85350000 85100000.0
#     # 85350000 - 84100000 = 1250000
#     #msec=84100000
#     n=gfile.split('_')
# 
#     cstart =  obs_end_dt - dt.timedelta(microseconds=msec)
#     st = cstart.strftime('%H%M%S%f')[0:7]   
# 
#     LOG.info('LOOK %s %s %s %s'%(  obs_start_dt,gran_start_dt , obs_end_dt , gran_end_dt))
#     tac = 'start adl %s guess %s %s diff %d  %s %s %s '%(n[3][1:],st, (st==n[3][1:]),int(int(st)-int(n[3][1:])), calc_gran_time , msec, emi )   
#      
#     return tac


def move_products_to_work_directory(work_dir, product_ids,optional=None):
    """ Checks that proper products were produced.
    If product h5 file was produced blob and lutasc files are deleted
    If product was not produced files are left alone
    """
    dest = os.path.dirname(work_dir)

    chunks=0
    option_chunks=0
    
    #once=False

    for product_id in product_ids:
        files = glob.glob(os.path.join(work_dir, '%s*.h5' % product_id))
        if chunks == 0 :
            chunks=len(files)
        for sdrproduct in files:
            if sdrproduct.find("_d1958") != -1:
                LOG.warn("IGNORE File with bad year: " + sdrproduct)
            else:
                new_name = os.path.basename(
                    sdrproduct.replace('_cspu_', '_cspp_'))
                new_name = os.path.basename(
                    new_name.replace('_pop', '_dev'))
                
                new_name = os.path.basename(
                    new_name.replace('_ops', '_dev'))
                
                 
                shutil.move(sdrproduct, dest, new_name)
#                 if once :
#                     once = False
#                     tac=eval_start_and_end(work_dir ,os.path.basename(sdrproduct))
# 
#                     LOG.info('Product %s  %s' % (os.path.basename(sdrproduct),tac))
         

          
    if optional is not  None:  
        for opt in optional:    
            files = glob.glob(os.path.join(work_dir, '%s*.h5' % opt))
            if option_chunks == 0:
                option_chunks =  len(files)
            for sdrproduct in files:
    
                new_name = os.path.basename(
                        sdrproduct.replace('_cspu_', '_cspp_'))
                new_name = os.path.basename(
                        new_name.replace('_pop', '_dev'))
                    
                new_name = os.path.basename(
                        new_name.replace('_ops', '_dev'))
                     
                shutil.move(sdrproduct, dest, new_name)
                LOG.debug('Product %s ' % sdrproduct)


        if len(files) < 1:
            LOG.debug('Product %s missing. %d products produced.' %
                      (product_id, len(files)))

            
    LOG.info('%s %s'%( chunks , option_chunks))

    return chunks , option_chunks


def gran_delta(gran1, gran2, expected_delta):
    """
    """
    g_id_1 = int(gran1['N_Granule_ID'][3:])
    g_id_2 = int(gran2['N_Granule_ID'][3:])

    g_delta = ((g_id_1 - g_id_2 + 5) / expected_delta)

    if g_delta == 0:
        g_delta = 1

    LOG.debug('grannum %d %d delta %d' % (g_id_1, g_id_2, g_delta))
    return g_delta


def trim_the_granules(granules_to_process, expected_delta, trim=None):
    """
    process the trim option
    Remove granules at beginning and or ending of strean,  return the trimmed list
    """
    if not trim:
        return granules_to_process

    LOG.info("Trim %d %d Number input granules %d" %
             (trim[0], trim[1], len(granules_to_process)))

    trim_begin = trim[0]
    trim_end = trim[1]
    total_gran = len(granules_to_process)

    LOG.info('Process between gran %d to %d in sequence of %d number to trim %d' %
             (trim_begin, (total_gran - trim_end)+1, total_gran, (trim_begin + trim_end) ))
    
    if (total_gran - trim_end) <= 0 :
        LOG.warn('Trim end would remove all granules, ignored')
        trim_end = 0

    if total_gran == 0:
        return granules_to_process

    LOG.debug('Before %d' % len(granules_to_process))
    tn = 0
    for gran1 in granules_to_process:
        LOG.debug('%2d id %s' % (tn, gran1['N_Granule_ID']))
        tn += 1

    trimmed_list = []
    num_to_trim = trim_begin
    granules = granules_to_process
    # not just a slice because missing grnaules should be counted
    prev = granules[0]
    for gran1 in granules[1:]:

        if num_to_trim > 0:
            #delta = gran_delta(gran2, gran1, expected_delta)
            delta = gran_delta(gran1, prev, expected_delta)

            LOG.debug(
                'Forward Trim %s %d delta %d granules missing %d ' % (prev['N_Granule_ID'], num_to_trim, delta, delta - 1))
            num_to_trim -= delta
        else:
            trimmed_list.insert(0, prev)

        prev = gran1
    # add last one
    trimmed_list.insert(0, prev)

    granules_to_process = []
    num_to_trim = trim_end
    gran2 = trimmed_list[0]

    LOG.debug('Middle %d' % len(trimmed_list))
    for gran1 in trimmed_list:
        LOG.debug('id %s' % gran1['N_Granule_ID'])

    for gran1 in trimmed_list[1:]:

        if num_to_trim > 0:
            delta = gran_delta(gran2, gran1, expected_delta)

            LOG.debug(
                'Backward Trim %s %d delta %d granules missing %d ' % (gran1['N_Granule_ID'], num_to_trim, delta, delta - 1))

            num_to_trim -= delta
        else:
            granules_to_process.insert(0, gran2)

        gran2 = gran1
    granules_to_process.insert(0, gran2)

    LOG.info('Number granules after trim %d' % len(granules_to_process))

    for gran1 in granules_to_process:
        LOG.info('id %s' % gran1['N_Granule_ID'])
    return granules_to_process


ANCILLARY_SUB_DIR = "linked_data"


def generate_sdr_xml(work_dir, gran, controller_name, controller_template):
    name = gran['N_Granule_ID']
    # fnxml = 'sdr_viirs_%s.xml' % name
    fnxml = controller_name % name
    #LOG.info('writing XML file %r' % fnxml)
    # fpxml = file(os.path.join(work_dir, fnxml), 'wt')

    fpxml = open(os.path.join(work_dir, fnxml), 'wt')
    # fpxml.write(XML_TMPL_VIIRS_SDR % gran)
    fpxml.write(controller_template % gran)
    status_line('Created ADL controller XML %s for %s' %
                (fnxml, gran['N_Granule_ID']))

    return fnxml


def setup_multi_process(work_dir, granules_to_process, controller_names, controller_templates, additional_keys=dict(),
                        compress=False,
                        cleanup=True,accuracyMode='reduced'):
    """
    put together the arguments needed to allow ADL and ancillary to run
    create xml for ADL controllers
    create processing directories
    create arguments to pass to multiplrocessing
    remove any granules already processed

    """
    argument_dictionaries = []
    chunk_size=int(additional_keys['CHUNK_SIZE'])
    #chunk_size=1
    if chunk_size <= 0:
        chunk_size=1
        
    chunks = [granules_to_process[x:x+chunk_size] for x in range(0, len(granules_to_process), chunk_size)]

    spccrft=additional_keys['SPCRFT']
    instrument=additional_keys['INSTRUMENT']

    #LOG.info(chunks)
    skip = []
    for granule_chunk  in chunks:
        gran = granule_chunk[0]
        gran_id = gran['N_Granule_ID']

        granule_output_dir = os.path.join(work_dir, gran_id)
        try:
            LOG.info('%s N_Granule_ID %s chunk_size %d  mkdir %s' %
                     (instrument,gran_id,chunk_size, granule_output_dir))
            os.mkdir(granule_output_dir)

            LOG.debug(pp.pformat(gran))

            xmls = []
            name = 0
           
                
            for template in controller_templates:
                sdr_xml = generate_sdr_xml2(
                    granule_output_dir, granule_chunk, controller_names[name], template,instrument,spccrft)
                full_sdr_xml = os.path.join(granule_output_dir, sdr_xml)
                xmls.append(full_sdr_xml)
                name += 1

            additional_envs = dict(
                COMPRESS=str(compress),
                N_Granule_ID=gran_id,
                XML_FILE=":".join(xmls),
                GRANULE_OUTPUT_DIR=granule_output_dir,
                ADL_HDF_OUTPUT_PATH=granule_output_dir,
                WORK_DIR=work_dir,
                # LINKED_ANCILLARY=anc_dir,
                ADL_HOME=ADL_HOME,
                # CSPP_RT_ANC_TILE_PATH=CSPP_RT_ANC_TILE_PATH,
                CLEANUP=str(cleanup),
                ACCURACYMODE=accuracyMode,
                DCONFIG=ADL_VARS['DCONFIG'],
                ACTUAL=str(len(granule_chunk))
            )

            additional_envs.update(additional_keys)
            argument_dictionaries.append(additional_envs)

        except Exception as e:
            LOG.error(traceback.format_exc())
            exc_type, exc_obj, exc_tb = sys.exc_info()

            LOG.info(str(exc_type))
            LOG.info("Granule %s Skipped,  Already processed" % gran_id)
            skip.append(gran)

    for gran in skip:
        granules_to_process.remove(gran)

    return argument_dictionaries, granules_to_process


def start_out_good(in_total_granules):
    """
    initialize the good count
    """
    global good_granules
    global complete_granules
    global total_granules

    global good_lock

    good_lock = Lock()
    good_granules = Value('i', 0)
    total_granules = Value('i', in_total_granules)
    complete_granules = Value('i', 0)


def report_gran_completion( stat,chunks,chunk_size):
    """
    Record that the granule has been processed and if it worked
    """
    global good_granules
    global complete_granules
    global total_granules
    global good_lock
    good_lock.acquire()
    ichunks = int(chunks)
    complete_granules.value += int(chunk_size)   
    good_granules.value += 1*ichunks
        
    good_lock.release()
    
    complete = complete_granules.value
    if complete_granules.value > total_granules.value :
        complete = total_granules.value

    status_line('%s %d out of %d granules processed, %d successfully' % (
        stat, complete, total_granules.value, good_granules.value))

    return good_granules


def number_good_granules():
    global good_granules
    if 'good_granules' in globals():
        return good_granules.value
    else:
        return 0


def linux_ram():
    """Returns the RAM of a linux system"""
    # totalMemory = os.popen("free -m").readlines()[1].split()[3]

    meminfo = open('/proc/meminfo', 'r').read()

    total_mem_kb_str, = re.findall(r'Inactive:\s+(\d+)\skB', meminfo)
    inactive = .75 * int(total_mem_kb_str) / 1024  # convert to MB

    total_mem_kb_str, = re.findall(r'MemFree:\s+(\d+)\skB', meminfo)
    free = int(total_mem_kb_str) / 1024  # convert to MB

    return int(inactive + free)


def check_number_processors(requested):
    number_available = cpu_count()

    nprocs = requested
    if int(nprocs) > number_available:
        LOG.warning("More processors requested %s than available %d" %
                    (nprocs, number_available))
        nprocs = number_available - 1
        if nprocs <= 0:
            nprocs = 1
    return nprocs


def check_memory(nprocs, gig_recomended):
    """
    Check the amount of memory availble against what is needed
    """
    free_gig_mem = linux_ram() / 1024
    gig_required = int(nprocs) * gig_recomended

    LOG.debug("Memory report: %d gig available  %d recommended for %d processors" % (
        int(free_gig_mem), int(gig_required), int(nprocs)))

    if gig_required > free_gig_mem:
        LOG.warn("Potential memory shortage: %d gig available  %d recommended for %d processors" % (
            int(free_gig_mem), int(gig_required), int(nprocs)))
        LOG.warn("Reduce number -p # or add memory")


def build_workspace(work_dir):
    "create work directory and subdirectories"
    if not os.path.isdir(work_dir):
        LOG.info('creating directory %s' % work_dir)
        os.makedirs(work_dir)
    log_dir = os.path.join(work_dir, 'log')
    if not os.path.isdir(log_dir):
        LOG.debug('creating directory %s' % log_dir)
        os.makedirs(log_dir)
    # anc_dir = os.path.join(work_dir, 'linked_data')
    # if not os.path.isdir(anc_dir):
    #     LOG.debug('creating directory %s' % anc_dir)
    #     os.makedirs(anc_dir)
    return log_dir


def check_products(work_dir, gran_id, product_shortnames,chunk_size=1):
    """ Checks that proper products were produced.
    If product h5 file was produced blob and lutasc files are deleted
    If product was not produced files are left alone
    """
    # look for all products that should have been produced
    # convert the products to H5

    LOG.info("Check for Granule: " + gran_id + " products.")
    start_time = time.time()
    LOG.debug("<< check_for_products  >>")

    # all = glob.glob(os.path.join(work_dir, '*'))

    problem = False
    total = 0
    good = 0
    for short_name in sorted(product_shortnames):
        prod_id = SHORTNAME_2_PRODUCTID[short_name]
        LOG.debug('look %s' % os.path.join(work_dir, '%s*.h5' % prod_id))
        gran_list = glob.glob(os.path.join(work_dir, '%s*' % prod_id))

        if len(gran_list) < 1:
            try:
                LOG.warn(
                    "Problem : " + gran_id + " " + short_name + " No product produced. " + prod_id)

                problem = True
            except KeyError:
                LOG.warn("Short name not in product map:" + short_name)
        else:
            for sdr_name in gran_list:
                try:
                    total = total + 1

                    LOG.debug("Product: " + sdr_name + " produced.")

                    if sdr_name.find("_d1958") != -1:
                        LOG.warn("DELETE File with bad year: " + sdr_name)
                        if os.path.exists(sdr_name):
                            os.remove(sdr_name)
                        problem = True
                    else:
                        good += 1

                except KeyError:
                    LOG.error(
                        "Key error on blob property " + 'ObservedStartTime' + " or " + short_name)

    LOG.info(str(good) + " out of " + str(total) +
             " products produced for granule " + gran_id)

    end_time = time.time()
    elapsed_time = end_time - start_time
    LOG.debug("<< Check Products Elapsed Time: %f >>" % elapsed_time)

    return problem

def abs_filenames(filenames):
    """
    :param filenames:
    :return:
    """

    newnames = []
    for name in filenames:
        if not os.path.exists(name):
            raise FileNotFoundError('file %s does not exist.' % name)

        newnames.append(os.path.abspath(name))
    return newnames


def cleanup_work_dir(work_dir,force_all=False,spcrft='NPP'):
    "upon successful run, clean out work directory"
    
    remaining_files_and_directories = glob.glob(os.path.join(work_dir, '%s*' % spcrft))

    for r in remaining_files_and_directories:
        shutil.rmtree(r)
    LOG.debug('remaining_files_and_directories %d' % len(remaining_files_and_directories))
    if len(remaining_files_and_directories) != 0 or force_all is True:
        dir_to_go = os.path.join(work_dir, "*.log")
        if os.path.exists(dir_to_go):
            shutil.rmtree(dir_to_go)
        dir_to_go = os.path.join(work_dir, "db")
        if os.path.exists(dir_to_go):
            shutil.rmtree(dir_to_go)
        dir_to_go = os.path.join(work_dir, "msd")
        if os.path.exists(dir_to_go):
            shutil.rmtree(dir_to_go)   
        dir_to_go = os.path.join(work_dir, "log")
        if os.path.exists(dir_to_go):
            shutil.rmtree(dir_to_go)                  

        #remove_inputs(work_dir)
         
    tlef = glob.glob(os.path.join(work_dir,'TLE*.*'))
    for tle in tlef: 
        os.remove(tle)
        
    tlef = glob.glob(os.path.join(work_dir,'*.msd'))
    for tle in tlef: 
        os.remove(tle)
     
    tin = glob.glob(os.path.join(work_dir,'TLINSET'))
    for tle in tin: 
        os.remove(tle)

        
    tlef = glob.glob(os.path.join(work_dir,'CrIS-SDR-ENGPKT-BACKUP-AUX*.*'))
    for tle in tlef:
        os.remove(tle)
    
    tlef = glob.glob(os.path.join(work_dir,'CrIS-FS-SDR-ENGPKT-BACKUP*.*'))
    for tle in tlef:
        os.remove(tle)

    

